package com.amazon.aws.buildersflash.schoolmailsearch;

import java.io.IOException;
import java.io.InputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.mail.BodyPart;
import javax.mail.MessagingException;
import javax.mail.Part;
import javax.mail.Session;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.internet.MimePart;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.text.PDFTextStripper;
import org.elasticsearch.action.index.IndexRequest;
import org.elasticsearch.action.index.IndexResponse;
import org.elasticsearch.client.RequestOptions;

import com.amazonaws.services.lambda.runtime.Context;
import com.amazonaws.services.lambda.runtime.events.S3Event;
import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.AmazonS3ClientBuilder;
import com.amazonaws.services.s3.event.S3EventNotification.S3Entity;
import com.amazonaws.services.s3.event.S3EventNotification.S3EventNotificationRecord;
import com.amazonaws.services.s3.model.S3Object;

public class AddDocumentFunction extends AbstractFunction {

    private static final AmazonS3 S3 = AmazonS3ClientBuilder.defaultClient();

    public void handle(S3Event event, Context context) {

        for (S3EventNotificationRecord record : event.getRecords()) {
            // S3 オブジェクトを取得
            S3Entity s3Entity = record.getS3();
            S3Object s3Object = S3.getObject(s3Entity.getBucket().getName(), s3Entity.getObject().getKey());

            // S3 オブジェクトの内容を取得しつつ、検索インデックスを作成
            try (InputStream is = s3Object.getObjectContent()) {
                addDocument(is);
            } catch (IOException | MessagingException e) {
                e.printStackTrace();
                continue;
            }
        }
    }

    private void addDocument(InputStream is) throws MessagingException, IOException {
        // InputStream データを MIME 形式として扱う
        Session session = Session.getDefaultInstance(new Properties());
        MimeMessage mimeMessage = new MimeMessage(session, is);
        // MimePart として検索インデックスを作る
        addDocumentAsMimePart(mimeMessage, mimeMessage.getSubject(), mimeMessage.getSentDate());
    }

    private void addDocumentAsMimePart(MimePart mimeMessage, String subject, Date sentDate)
            throws MessagingException, IOException {
        if (mimeMessage.isMimeType("multipart/*")) {
            // multipart の場合は、それぞれの part ごとに検索インデックスを作る
            MimeMultipart multipart = (MimeMultipart) mimeMessage.getContent();
            System.out.println("partCount: " + multipart.getCount());
            for (int i = 0, count = multipart.getCount(); i < count; i++) {
                BodyPart bodyPart = multipart.getBodyPart(i);
                try {
                    addDocumentAsPart(bodyPart, subject, sentDate);
                } catch (Exception e) {
                    e.printStackTrace();
                    continue;
                }
            }
        } else {
            // それ以外の場合は、検索インデックスを 1 つだけ作る
            addDocumentAsPart(mimeMessage, subject, sentDate);
        }
    }

    private void addDocumentAsPart(Part part, String subject, Date sentDate) throws MessagingException, IOException {
        String contentType = part.getContentType();
        System.out.println("!Content-Type: " + contentType);
        if (part.isMimeType("text/html")) {
            // text/plain の方でインデックスが作られることを見越し、text/html 形式は対象外
            return;
        }
        if (part.isMimeType("multipart/*")) {
            // part がさらに multipart だった場合は、再帰的に呼び出す
            addDocumentAsMimePart((MimePart) part, subject, sentDate);
            return;
        }

        String content;
        if (part.isMimeType("text/*")) {
            // text の場合は、そのまま
            content = getContentAsText(part);
        } else if (part.isMimeType("application/pdf")) {
            // PDF の場合は、テキストを抽出する
            content = getContentAsPdf(part);
        } else {
            // それ以外はスキップ
            System.out.println("Unsupported Content Type: " + contentType);
            return;
        }
        System.out.println("content: " + content);

        // 検索インデックスの作成
        Map<String, String> document = new HashMap<>();
        if (StringUtils.isNotEmpty(subject)) {
            document.put("subject", subject);
        }
        document.put("content", content);
        document.put("sentDate", String.valueOf(sentDate.getTime()));

        // 検索インデックスを追加
        IndexRequest request = new IndexRequest(INDEX_NAME).source(document);
        try {
            IndexResponse response = ES_CLIENT.index(request, RequestOptions.DEFAULT);
            System.out.println(response);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private static final Pattern CONTENT_TYPE_REGEXP = Pattern.compile(".*charset=\"(.*)\"");

    private String getContentAsText(Part part) throws MessagingException, IOException {
        // charset の取得
        String charset = "UTF-8";
        Matcher matcher = CONTENT_TYPE_REGEXP.matcher(part.getContentType());
        if (matcher.find()) {
            charset = matcher.group(1);
        }
        System.out.println("charset: " + charset);

        // テキストを返す
        try (InputStream messageInputStream = part.getInputStream()) {
            return IOUtils.toString(messageInputStream, charset);
        }
    }

    private String getContentAsPdf(Part part) throws MessagingException, IOException {
        try (InputStream is = part.getInputStream()) {
            PDFTextStripper textStripper = new PDFTextStripper();

            // PDF の全ページからテキストを抽出
            PDDocument document = PDDocument.load(is);
            textStripper.setStartPage(1);
            textStripper.setEndPage(document.getNumberOfPages());

            return textStripper.getText(document);
        }
    }
}
