package com.amazon.aws.buildersflash.schoolmailsearch;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.common.text.Text;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightField;

import com.amazonaws.services.lambda.runtime.Context;
import com.amazonaws.services.lambda.runtime.events.APIGatewayProxyRequestEvent;
import com.amazonaws.services.lambda.runtime.events.APIGatewayProxyResponseEvent;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

public class SearchDocumentFunction extends AbstractFunction {

    private static final SimpleDateFormat DATE_FORMAT;
    static {
        DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("Asia/Tokyo"));
    }

    public APIGatewayProxyResponseEvent handle(APIGatewayProxyRequestEvent event, Context context) {

        // パラメータの取得
        Map<String, String> queryStringParameters = event.getQueryStringParameters();

        String queryString = queryStringParameters != null ? queryStringParameters.get("query") : null;
        if (StringUtils.isEmpty(queryString)) {
            return createResponse(Collections.emptyList(), queryStringParameters);
        }

        // 日本語検索のため "" で囲む
        MultiMatchQueryBuilder multiMatchQueryBuilder =
                QueryBuilders.multiMatchQuery(String.format("\"%s\"", queryString), "subject", "content");

        // content に対してハイライターを設定
        HighlightBuilder highlightBuilder = new HighlightBuilder();
        HighlightBuilder.Field highlightContent = new HighlightBuilder.Field("content");
        highlightBuilder.field(highlightContent);

        // 検索条件を作成
        SearchSourceBuilder searchSourceBuilder =
                SearchSourceBuilder.searchSource().query(multiMatchQueryBuilder).highlighter(highlightBuilder);

        List<Mail> responses = new ArrayList<>();
        try {
            // 検索
            SearchResponse searchResponse =
                    ES_CLIENT.search(new SearchRequest(INDEX_NAME).source(searchSourceBuilder), RequestOptions.DEFAULT);
            
            // 検索結果を Mail オブジェクトに変換
            for (SearchHit hit : searchResponse.getHits()) {
                List<String> highlightedContent = new ArrayList<>();
                HighlightField highlightFieldContent = hit.getHighlightFields().get("content");
                if (highlightFieldContent != null) {
                    for (Text text : highlightFieldContent.fragments()) {
                        highlightedContent.add(text.toString());
                    }
                }

                Mail mail = new Mail();

                Map<String, Object> hitMap = hit.getSourceAsMap();
                mail.setSubject((String) hitMap.get("subject"));
                mail.setContent((String) hitMap.get("content"));
                mail.setHighlightedContent(highlightedContent);
                mail.setSentDate(new Date(Long.valueOf((String) hitMap.get("sentDate"))));

                responses.add(mail);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        // 検索結果を返す
        return createResponse(responses, queryStringParameters);
    }

    private APIGatewayProxyResponseEvent createResponse(List<Mail> mails, Map<String, String> queryStringParameters) {
        // API Gateway の Lambda プロキシ統合のルールに沿ったレスポンスを返す
        APIGatewayProxyResponseEvent response = new APIGatewayProxyResponseEvent();
        response.setStatusCode(200);
        response.setHeaders(Collections.singletonMap("Content-Type", "application/json"));
        if (mails != null) {
            try {
                // Mail の List を JSON 文字列に変換
                response.setBody(new ObjectMapper().writeValueAsString(mails));
            } catch (JsonProcessingException e) {
                throw new RuntimeException(e);
            }
        }
        return response;
    }
}
